@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_caching
     ~~~~~~~~~~~~~
@@ -13,23 +12,36 @@ import functools
 import hashlib
 import inspect
 import logging
-import string
 import uuid
 import warnings
 from collections import OrderedDict
+from typing import Any
+from typing import Callable
+from typing import Optional
+from typing import Tuple
+from typing import Union
 
-from flask import current_app, request, url_for, Flask
+from flask import current_app
+from flask import Flask
+from flask import request
+from flask import Response
+from flask import url_for
+from markupsafe import Markup
 from werkzeug.utils import import_string
+
 from flask_caching.backends.base import BaseCache
 from flask_caching.backends.simplecache import SimpleCache
-from markupsafe import Markup
-from typing import Any, Callable, List, Optional, Tuple, Union
+from flask_caching.utils import function_namespace
+from flask_caching.utils import get_arg_default
+from flask_caching.utils import get_arg_names
+from flask_caching.utils import get_id
+from flask_caching.utils import make_template_fragment_key  # noqa: F401
+from flask_caching.utils import wants_args
 
-__version__ = "1.10.1"
+__version__ = "2.0.1"
 
 logger = logging.getLogger(__name__)
 
-TEMPLATE_FRAGMENT_KEY_TEMPLATE = "_template_fragment_cache_%s%s"
 SUPPORTED_HASH_FUNCTIONS = [
     hashlib.sha1,
     hashlib.sha224,
@@ -39,112 +51,21 @@ SUPPORTED_HASH_FUNCTIONS = [
     hashlib.md5,
 ]
 
-# Used to remove control characters and whitespace from cache keys.
-valid_chars = set(string.ascii_letters + string.digits + "_.")
-delchars = "".join(c for c in map(chr, range(256)) if c not in valid_chars)
-null_control = (dict((k, None) for k in delchars),)
 
-
-def wants_args(f: Callable) -> bool:
-    """Check if the function wants any arguments"""
-
-    argspec = inspect.getfullargspec(f)
-
-    return bool(argspec.args or argspec.varargs or argspec.varkw)
-
-
-def get_arg_names(f: Callable) -> List[str]:
-    """Return arguments of function
-
-    :param f:
-    :return: String list of arguments
+class CachedResponse(Response):
     """
-    sig = inspect.signature(f)
-    return [
-        parameter.name
-        for parameter in sig.parameters.values()
-        if parameter.kind == parameter.POSITIONAL_OR_KEYWORD
-    ]
+    views wraped by @cached can return this (which inherits from flask.Response)
+    to override the cache TTL dynamically
+    """
+
+    timeout = None
+
+    def __init__(self, response, timeout):
+        self.__dict__ = response.__dict__
+        self.timeout = timeout
 
 
-def get_arg_default(f: Callable, position: int):
-    sig = inspect.signature(f)
-    arg = list(sig.parameters.values())[position]
-    arg_def = arg.default
-    return arg_def if arg_def != inspect.Parameter.empty else None
-
-
-def get_id(obj):
-    return getattr(obj, "__caching_id__", repr)(obj)
-
-
-def function_namespace(f, args=None):
-    """Attempts to returns unique namespace for function"""
-    m_args = get_arg_names(f)
-
-    instance_token = None
-
-    instance_self = getattr(f, "__self__", None)
-
-    if instance_self and not inspect.isclass(instance_self):
-        instance_token = get_id(f.__self__)
-    elif m_args and m_args[0] == "self" and args:
-        instance_token = get_id(args[0])
-
-    module = f.__module__
-
-    if m_args and m_args[0] == "cls" and not inspect.isclass(args[0]):
-        raise ValueError(
-            "When using `delete_memoized` on a "
-            "`@classmethod` you must provide the "
-            "class as the first argument"
-        )
-
-    if hasattr(f, "__qualname__"):
-        name = f.__qualname__
-    else:
-        klass = getattr(f, "__self__", None)
-
-        if klass and not inspect.isclass(klass):
-            klass = klass.__class__
-
-        if not klass:
-            klass = getattr(f, "im_class", None)
-
-        if not klass:
-            if m_args and args:
-                if m_args[0] == "self":
-                    klass = args[0].__class__
-                elif m_args[0] == "cls":
-                    klass = args[0]
-
-        if klass:
-            name = klass.__name__ + "." + f.__name__
-        else:
-            name = f.__name__
-
-    ns = ".".join((module, name))
-    ns = ns.translate(*null_control)
-
-    if instance_token:
-        ins = ".".join((module, name, instance_token))
-        ins = ins.translate(*null_control)
-    else:
-        ins = None
-
-    return ns, ins
-
-
-def make_template_fragment_key(
-    fragment_name: str, vary_on: List[str] = []
-) -> str:
-    """Make a cache key for a specific fragment name."""
-    if vary_on:
-        fragment_name = "%s_" % fragment_name
-    return TEMPLATE_FRAGMENT_KEY_TEMPLATE % (fragment_name, "_".join(vary_on))
-
-
-class Cache(object):
+class Cache:
     """This class is used to control the cache objects."""
 
     def __init__(
@@ -194,18 +115,18 @@ class Cache(object):
         config.setdefault("CACHE_NO_NULL_WARNING", False)
         config.setdefault("CACHE_SOURCE_CHECK", False)
 
-        if (
-            config["CACHE_TYPE"] == "null"
-            and not config["CACHE_NO_NULL_WARNING"]
-        ):
+        if config["CACHE_TYPE"] == "null" and not config["CACHE_NO_NULL_WARNING"]:
             warnings.warn(
                 "Flask-Caching: CACHE_TYPE is set to null, "
                 "caching is effectively disabled."
             )
 
-        if config["CACHE_TYPE"] == "filesystem" and config["CACHE_DIR"] is None:
+        if (
+            config["CACHE_TYPE"] in ["filesystem", "FileSystemCache"]
+            and config["CACHE_DIR"] is None
+        ):
             warnings.warn(
-                "Flask-Caching: CACHE_TYPE is set to filesystem but no "
+                f"Flask-Caching: CACHE_TYPE is set to {config['CACHE_TYPE']} but no "
                 "CACHE_DIR is set."
             )
 
@@ -253,6 +174,12 @@ class Cache(object):
         )
         self.app = app
 
+    def _call_fn(self, fn, *args, **kwargs):
+        ensure_sync = getattr(self.app, "ensure_sync", None)
+        if ensure_sync is not None:
+            return ensure_sync(fn)(*args, **kwargs)
+        return fn(*args, **kwargs)
+
     @property
     def cache(self) -> SimpleCache:
         app = current_app or self.app
@@ -262,6 +189,10 @@ class Cache(object):
         """Proxy function for internal cache object."""
         return self.cache.get(*args, **kwargs)
 
+    def has(self, *args, **kwargs) -> bool:
+        """Proxy function for internal cache object."""
+        return self.cache.has(*args, **kwargs)
+
     def set(self, *args, **kwargs) -> bool:
         """Proxy function for internal cache object."""
         return self.cache.set(*args, **kwargs)
@@ -382,8 +313,8 @@ class Cache(object):
                               renewal of cached functions.
 
         :param response_filter: Default None. If not None, the callable is
-                                invoked after the cached funtion evaluation,
-                                and is given one arguement, the response
+                                invoked after the cached function evaluation,
+                                and is given one argument, the response
                                 content. If the callable returns False, the
                                 content will not be cached. Useful to prevent
                                 caching of code 500 responses.
@@ -422,7 +353,7 @@ class Cache(object):
             def decorated_function(*args, **kwargs):
                 #: Bypass the cache entirely.
                 if self._bypass_cache(unless, f, *args, **kwargs):
-                    return f(*args, **kwargs)
+                    return self._call_fn(f, *args, **kwargs)
 
                 nonlocal source_check
                 if source_check is None:
@@ -432,9 +363,7 @@ class Cache(object):
                     if make_cache_key is not None and callable(make_cache_key):
                         cache_key = make_cache_key(*args, **kwargs)
                     else:
-                        cache_key = _make_cache_key(
-                            args, kwargs, use_request=True
-                        )
+                        cache_key = _make_cache_key(args, kwargs, use_request=True)
 
                     if (
                         callable(forced_update)
@@ -469,24 +398,27 @@ class Cache(object):
                     if self.app.debug:
                         raise
                     logger.exception("Exception possibly due to cache backend.")
-                    return f(*args, **kwargs)
+                    return self._call_fn(f, *args, **kwargs)
 
                 if not found:
-                    rv = f(*args, **kwargs)
+                    rv = self._call_fn(f, *args, **kwargs)
+                    if inspect.isgenerator(rv):
+                        rv = [val for val in rv]
 
                     if response_filter is None or response_filter(rv):
+                        cache_timeout = decorated_function.cache_timeout
+                        if isinstance(rv, CachedResponse):
+                            cache_timeout = rv.timeout or cache_timeout
                         try:
                             self.cache.set(
                                 cache_key,
                                 rv,
-                                timeout=decorated_function.cache_timeout,
+                                timeout=cache_timeout,
                             )
                         except Exception:
                             if self.app.debug:
                                 raise
-                            logger.exception(
-                                "Exception possibly due to cache backend."
-                            )
+                            logger.exception("Exception possibly due to cache backend.")
                 return rv
 
             def default_make_cache_key(*args, **kwargs):
@@ -520,7 +452,7 @@ class Cache(object):
                 # provided.
 
                 args_as_sorted_tuple = tuple(
-                    sorted((pair for pair in request.args.items(multi=True)))
+                    sorted(pair for pair in request.args.items(multi=True))
                 )
                 # ... now hash the sorted (key, value) tuple so it can be
                 # used as a key for cache. Turn them into bytes so that the
@@ -551,17 +483,13 @@ class Cache(object):
                         if use_request:
                             cache_key = key_prefix % request.path
                         else:
-                            cache_key = key_prefix % url_for(
-                                f.__name__, **kwargs
-                            )
+                            cache_key = key_prefix % url_for(f.__name__, **kwargs)
                     else:
                         cache_key = key_prefix
 
                 if source_check and callable(f):
                     func_source_code = inspect.getsource(f)
-                    func_source_hash = hash_method(
-                        func_source_code.encode("utf-8")
-                    )
+                    func_source_hash = hash_method(func_source_code.encode("utf-8"))
                     func_source_hash = str(func_source_hash.hexdigest())
 
                     cache_key += func_source_hash
@@ -653,7 +581,7 @@ class Cache(object):
 
     def _memoize_make_cache_key(
         self,
-        make_name: None = None,
+        make_name: Optional[Callable] = None,
         timeout: Optional[Callable] = None,
         forced_update: bool = False,
         hash_method: Callable = hashlib.md5,
@@ -684,7 +612,7 @@ class Cache(object):
             else:
                 keyargs, keykwargs = args, kwargs
 
-            updated = u"{0}{1}{2}".format(altfname, keyargs, keykwargs)
+            updated = f"{altfname}{keyargs}{keykwargs}"
 
             cache_key = hash_method()
             cache_key.update(updated.encode("utf-8"))
@@ -707,7 +635,7 @@ class Cache(object):
         #: Inspect the arguments to the function
         #: This allows the memoization to be the same
         #: whether the function was called with
-        #: 1, b=2 is equivilant to a=1, b=2, etc.
+        #: 1, b=2 is equivalent to a=1, b=2, etc.
         new_args = []
         arg_num = 0
         args_to_ignore = kwargs.pop("args_to_ignore", None) or []
@@ -768,9 +696,7 @@ class Cache(object):
         return (
             tuple(new_args),
             OrderedDict(
-                sorted(
-                    (k, v) for k, v in kwargs.items() if k in kw_keys_remaining
-                )
+                sorted((k, v) for k, v in kwargs.items() if k in kw_keys_remaining)
             ),
         )
 
@@ -799,10 +725,10 @@ class Cache(object):
     def memoize(
         self,
         timeout: Optional[int] = None,
-        make_name: None = None,
-        unless: None = None,
+        make_name: Optional[Callable] = None,
+        unless: Optional[Callable] = None,
         forced_update: Optional[Callable] = None,
-        response_filter: None = None,
+        response_filter: Optional[Callable] = None,
         hash_method: Callable = hashlib.md5,
         cache_none: bool = False,
         source_check: Optional[bool] = None,
@@ -902,16 +828,14 @@ class Cache(object):
             def decorated_function(*args, **kwargs):
                 #: bypass cache
                 if self._bypass_cache(unless, f, *args, **kwargs):
-                    return f(*args, **kwargs)
+                    return self._call_fn(f, *args, **kwargs)
 
                 nonlocal source_check
                 if source_check is None:
                     source_check = self.source_check
 
                 try:
-                    cache_key = decorated_function.make_cache_key(
-                        f, *args, **kwargs
-                    )
+                    cache_key = decorated_function.make_cache_key(f, *args, **kwargs)
 
                     if (
                         callable(forced_update)
@@ -946,10 +870,12 @@ class Cache(object):
                     if self.app.debug:
                         raise
                     logger.exception("Exception possibly due to cache backend.")
-                    return f(*args, **kwargs)
+                    return self._call_fn(f, *args, **kwargs)
 
                 if not found:
-                    rv = f(*args, **kwargs)
+                    rv = self._call_fn(f, *args, **kwargs)
+                    if inspect.isgenerator(rv):
+                        rv = [val for val in rv]
 
                     if response_filter is None or response_filter(rv):
                         try:
@@ -961,9 +887,7 @@ class Cache(object):
                         except Exception:
                             if self.app.debug:
                                 raise
-                            logger.exception(
-                                "Exception possibly due to cache backend."
-                            )
+                            logger.exception("Exception possibly due to cache backend.")
                 return rv
 
             decorated_function.uncached = f
